/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.engine.source;

import java.io.Serializable;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.TimeZone;

import sync4j.framework.core.AlertCode;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.AbstractSyncSource;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.server.store.PersistentStoreException;

import sync4j.exchange.items.common.model.Item;
import sync4j.exchange.items.common.manager.ItemManager;
import sync4j.exchange.DataAccessException;

import sync4j.framework.security.Sync4jPrincipal;

import sync4j.server.config.Configuration;

/**
 *
 * This class implements common methods of<br>
 *  ExchangeCalendarSyncSource,
 *  ExchangeContactSyncSource,
 *  ExchangeNoteSyncSource, and
 *  ExchangeTaskSyncSource.
 *
 *
 * version $Id: ExchangeSyncSource.java,v 1.20 2006/02/13 13:39:36 fabius Exp $
 */
public abstract class ExchangeSyncSource
extends AbstractSyncSource
implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    protected static final String ITEM_ADD          = "A"                ;
    protected static final String ITEM_REMOVE       = "D"                ;
    protected static final String ITEM_UPDATE       = "U"                ;

    protected static final String ITEM_ADD_ERROR    = "AE"               ;
    protected static final String ITEM_REMOVE_ERROR = "DE"               ;
    protected static final String ITEM_UPDATE_ERROR = "UE"               ;

    protected static final String TYPE_VCARD        = "text/x-vcard"     ;
    protected static final String TYPE_ICAL         = "text/x-vcalendar" ;

    // -------------------------------------------------------------- Private data

    private   ItemManager  im                 = null ;

    private   Item[]    newItems              = null ;
    private   Item[]    deleteItems           = null ;
    private   Item[]    updateItems           = null ;

    private   ArrayList syncNewItems          = null ;
    private   ArrayList syncUpdateItems       = null ;
    private   ArrayList syncRemoveItems       = null ;

    private   ArrayList syncErrorNewItems     = null ;
    private   ArrayList syncErrorUpdateItems  = null ;
    private   ArrayList syncErrorRemoveItems  = null ;

    protected String    deviceTimeZoneDescr   = null ;
    protected TimeZone  deviceTimeZone        = null ;

    protected String    deviceCharset         = null  ;

    protected Logger    log                   = null  ;

    // -------------------------------------------------------------- Properties

     /**
      * Property exchange folder
      */
     protected String exchangeFolder;

     /** Getter for property exchange folder.
      * @return Value of property exchange folder.
      *
      */
     public String getExchangeFolder() {
         return exchangeFolder;
     }

     /** Setter for property exchange folder.
      * @param exchangeFolder New value of property exchange folder.
      *
      */
     public void setExchangeFolder(String exchangeFolder) {
         this.exchangeFolder = exchangeFolder;
     }

    /**
     * Property ExchangeServer host
     */
    private String host;

    /** Getter for property exchangeServerHost.
     * @return Value of property exchangeServerHost.
     *
     */
    public String getHost() {
        return host;
    }

    /** Setter for property exchangeServerHost.
     * @param host New value of property exchangeServerHost.
     *
     */
    public void setHost(String host) {
        this.host = host;
    }

    /**
     * Property ExchangeServer port
     */
    private int port = 80;

    /** Getter for property exchangeServerPort.
     * @return Value of property exchangeServerPort.
     *
     */
    public int getPort() {
        return port;
    }

    /** Setter for property exchangeServerPort.
     * @param port New value of property exchangeServerPort.
     *
     */
    public void setPort(int port) {
        this.port = port;
    }

    /**
     * Should the content be encoded?
     */
    private boolean encode = false;

    /** Getter for property encode.
     * @return Value of property encode.
     *
     */
    public boolean isEncode() {
        return encode;
    }

    /** Setter for property encode.
     * @param encode New value of property encode.
     *
     */
    public void setEncode(boolean encode) {
        this.encode = encode;
    }

    /**
     * Property schedule
     */
    private String schedule;

    /** Getter for property schedule.
     * @return Value of property schedule.
     *
     */
    public String getSchedule() {
        return schedule;
    }

    /** Setter for property schedule.
     * @param schedule New value of property schedule.
     *
     */
    public void setSchedule(String schedule) {
        this.schedule = schedule;
    }

    /**
     * Property scheduleEnabled
     */
    private boolean scheduleEnabled;

    /** Getter for property scheduleEnabled.
     * @return Value of property scheduleEnabled.
     *
     */
    public boolean isSchedulEnablede() {
        return scheduleEnabled;
    }

    /** Setter for property schedule.
     * @param scheduleEnabled New value of property schedule.
     *
     */
    public void setScheduleEnabled(boolean scheduleEnabled) {
        this.scheduleEnabled = scheduleEnabled;
    }

    /** Getter for property newItem.
     * @return Value of property newItems.
     *
     */
    public Item[] getNewItems()    {
        return this.newItems;
    }

    /** Getter for property updateItems.
     * @return Value of property updateItems.
     *
     */
    public Item[] getUpdateItems() {
        return this.updateItems;
    }

    /** Getter for property deleteItems.
     * @return Value of property deleteItems.
     *
     */
    public Item[] getDeleteItems() {
        return this.deleteItems;
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * @see SyncSource
     */
    public void beginSync(Principal principal, int syncMode)
    throws SyncSourceException {

        Item[]       localItems     = null ;
        Item[]       exchangeItems  = null ;

        int          principalId    = 0    ;
        String       username       = null ;
        String       credentials    = null ;

        log = Sync4jLogger.getLogger("source");

        syncNewItems          = new ArrayList();
        syncUpdateItems       = new ArrayList();
        syncRemoveItems       = new ArrayList();

        syncErrorNewItems     = new ArrayList();
        syncErrorUpdateItems  = new ArrayList();
        syncErrorRemoveItems  = new ArrayList();

        principalId = Integer.parseInt(((Sync4jPrincipal) principal).getId());
        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        try {
            String deviceId      = ((Sync4jPrincipal)principal).getDeviceId();
            Sync4jDevice device  = getDevice             (deviceId)  ;
            String timezone      = device.getTimeZone    (        )  ;
            if (device.getConvertDate ()) {
                if (timezone != null && timezone.length() > 0) {
                    deviceTimeZoneDescr = timezone;
                    deviceTimeZone = TimeZone.getTimeZone(deviceTimeZoneDescr);
                }
            }

            deviceCharset = device.getCharset();
        } catch (Exception e) {
            throw new SyncSourceException(e.getMessage());
        }

        try {

            im  = new ItemManager (this.host ,
                                   this.port);

            exchangeItems  = getExchangeItems  (username    ,
                                                credentials ,
                                                this.exchangeFolder,
                                                new String[0]) ;

            localItems = im.getLocalItems (this.sourceURI,
                                           principalId)  ;

            newItems     = findNewItems    (localItems, exchangeItems) ;
            updateItems  = findUpdateItems (localItems, exchangeItems) ;
            deleteItems  = findDeleteItems (localItems, exchangeItems) ;

            im.addItems (newItems,
                         this.sourceURI,
                         principalId);

            im.updateItems (updateItems,
                            this.sourceURI,
                            principalId);

            im.deleteItems (deleteItems,
                            this.sourceURI,
                            principalId);

            if (log.isLoggable(Level.INFO)) {
                log.info("Exchange SyncSource " + sourceURI + " - begineSync");
            }

        } catch (DataAccessException e) {
            throw new SyncSourceException("Error accessing items: " +
                                           e.getMessage(), e);
        }


    }

    /**
     * @see SyncSource
     */
    public void endSync(Principal principal)
    throws SyncSourceException {

        ItemManager  im                  = null  ;

        Item  []     exchItems           = null  ;
        String       id                  = null  ;

        int          principalId         = 0     ;
        String       username            = null  ;
        String       credentials         = null  ;

        principalId = Integer.parseInt(((Sync4jPrincipal) principal).getId());
        username    = ((Sync4jPrincipal) principal).getUsername()           ;
        credentials = ((Sync4jPrincipal) principal).getEncodedCredentials() ;

        try {

            int count = 0;

            im  = new ItemManager (this.host ,
                                   this.port );

            Item[] sni = (Item[]) syncNewItems.toArray(new Item[syncNewItems.size()]);

            Item[] sui = (Item[]) syncUpdateItems.toArray(new Item[syncUpdateItems.size()]);

            int k = sni.length + sui.length;

            //
            // upgrade only if items is added / updated in sync
            //
            if (k > 0) {

                String[] ids = new String [k];

                for (int i= 0, l = sni.length; i < l; i++) {
                    ids[count++] = sni[i].getId();
                }

                for (int i= 0, l = sui.length; i < l; i++) {
                    ids[count++] = sui[i].getId();
                }


                Item[] exchangeSyncItems  = getExchangeItems (username    ,
                                                              credentials ,
                                                              this.exchangeFolder,
                                                              ids) ;

                im.upgradeItems (exchangeSyncItems,
                                 this.sourceURI,
                                 principalId);

            }

            im.deleteItems ((Item[]) syncRemoveItems.toArray(new Item[syncRemoveItems.size()]),
                            this.sourceURI,
                            principalId);

            if (log.isLoggable(Level.INFO)) {
                log.info("Exchange SyncSource " + sourceURI + " - endSync");
            }

            if (log.isLoggable(Level.FINEST)) {

                log.finest("Items added: " + syncNewItems.size());

                Iterator i = syncNewItems.iterator();

                while(i.hasNext()) {
                    id = ((Item) i.next()).getId();
                }

                log.finest("Items updated: " + syncUpdateItems.size());

                i = syncUpdateItems.iterator();

                while(i.hasNext()) {
                    id = ((Item) i.next()).getId();
                }

                log.finest("Items removed: " + syncRemoveItems.size());

                i = syncRemoveItems.iterator();

                while(i.hasNext()) {
                    id = ((Item) i.next()).getId();
                }

            }

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "endSync", e);
            }

            throw new SyncSourceException("Error updating items:" +
                                          e.getMessage(), e);
        }

    }

   /**
     *
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @param ids
     *
     * @return items array
     *
     **/
    public abstract Item[] getExchangeItems (String username       ,
                                             String credentials    ,
                                             String exchangeFolder ,
                                             String[] ids)
    throws SyncSourceException;


   // ---------------------------------------------------------- Protected Methods

   /**
     * Return href (webdav resource name) of item
     *
     * @param id id of item to find
     * @param principal

     * @return href (webdav resource name) of find item,
     *         or <p>null</p> if item not found
     */
    protected String getHref(String id, int principal)
    throws SyncSourceException {

        String href = null;

        try {
            href = im.getHref(id, principal, this.sourceURI);
        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting href:" +
                                                  e.getMessage(), e);
        }

        return href;

    }

    /**
     * Add item to new items array list
     *
     * @param id
     * @param crc
     * @param lastUpdate
     *
     */
    protected void toSyncNewItems(String id, long crc, Date lastUpdate) {
        syncNewItems.add(new Item(id, lastUpdate, crc));
    }


    /**
     * Add item to update items array list
     *
     * @param id
     * @param crc
     * @param lastUpdate
     *
     */
    protected void toSyncUpdateItems(String id, long crc, Date lastUpdate) {
        syncUpdateItems.add(new Item(id, lastUpdate, crc));
    }


    /**
     * Add item to remove items array list
     *
     * @param id
     *
     */
    protected void toSyncRemoveItems(String id) {
        syncRemoveItems.add(new Item(id, null, 0));
    }

    /**
     * Add item to error new items array list
     *
     * @param id
     *
     */
    protected void toSyncErrorNewItems(String id) {
        syncErrorNewItems.add(id);
    }

    /**
     * Add item to error update items array list
     *
     * @param id
     *
     */
    protected void toSyncErrorUpdateItems(String id) {
        syncErrorUpdateItems.add(id);
    }

    /**
     * Add item to error remove items array list
     *
     * @param id
     *
     */
    protected void toSyncErrorRemoveItems(String id) {
        syncErrorRemoveItems.add(id);
    }

    /** Getter for property newItem ids.
      * @return Value of property newItems.
      *
      */
     protected String[] getNewItemIds() {

         String[] ids = null;

         if (newItems != null && newItems.length > 0) {

             int l = newItems.length;

             ids = new String[l];

             for (int i=0; i < l; i++) {
                ids[i] = newItems[i].getId();
             }

         } else {
             ids = new String[0];
         }

         return ids;
     }

     /** Getter for property updateItems ids.
      * @return Value of property updateItems.
      *
      */
     protected String[] getUpdateItemIds() {

         String[] ids = null;

         if (updateItems != null && updateItems.length > 0) {

             int l = updateItems.length;

             ids = new String[l];

             for (int i=0; i < l; i++) {
                ids[i] = updateItems[i].getId();
             }

         } else {
             ids = new String[0];
         }

         return ids;
     }

      /** Getter for property deleteItems ids.
       * @return Value of property updateItems.
       *
       */
      protected String[] getDeleteItemIds() {

          String[] ids = null;

          if (deleteItems != null && deleteItems.length > 0) {

              int l = deleteItems.length;

              ids = new String[l];

              for (int i=0; i < l; i++) {
                 ids[i] = deleteItems[i].getId();
              }

          } else {
              ids = new String[0];
          }

          return ids;
     }

    // ---------------------------------------------------------- Private Methods

    /**
     * Create new items array
     *
     * @param localItems
     * @param exchangeItems
     *
     * @return new items array
     */
    private Item[] findNewItems (Item[] localItems, Item[] exchangeItems) {

        ArrayList  newRows       = null ;
        Item       exchangeItem  = null ;

        int        count         = 0    ;

        newRows = new ArrayList();

        if (exchangeItems != null) {

            for (int i = 0, l = exchangeItems.length; i < l; i++) {

                exchangeItem = exchangeItems[i];

                if (!this.isItemInArray(localItems, exchangeItem)) {
                    newRows.add(exchangeItem);
                    count++;
                }

            }

        }

        return (Item[])newRows.toArray(new Item[count]) ;

    }

    /**
     * Create deleted items array
     *
     * @param localItems
     * @param exchangeItems
     *
     * @return deleted items array
     */
    private Item[] findDeleteItems (Item[] localItems, Item[] exchangeItems) {

        ArrayList  deleteRows       = null ;
        Item       localItem        = null ;

        int        count            = 0;

        deleteRows = new ArrayList();

        if (localItems != null) {

            for (int i=0, l = localItems.length; i < l; i++) {

               localItem = localItems[i];

               if (!this.isItemInArray(exchangeItems, localItem)) {
                   deleteRows.add(localItem);
                   count++;
               }

            }

        }

        return (Item[])deleteRows.toArray(new Item[count]) ;

    }

    /**
     * Create updated items array
     *
     * @param localItems
     * @param exchangeItems
     *
     * @return updated items array
     */
    private Item[] findUpdateItems (Item[] localItems, Item[] exchangeItems) {

        ArrayList  updateRows               = null ;

        Item       exchangeItem             = null ;
        Item       localItem                = null ;

        Date       excItemLastModified      = null ;
        Date       localItemLastModified    = null ;

        String     excItemLastModifiedStr   = null;
        String     localItemLastModifiedStr = null ;

        int        count                    = 0    ;

        updateRows = new ArrayList();


        if (exchangeItems != null) {

            for (int i = 0, l = exchangeItems.length; i < l; i++) {

                exchangeItem   = exchangeItems[i]                  ;
                localItem      = getItemById(localItems,
                                             exchangeItem.getId()) ;

                if (localItem != null) {

                     excItemLastModified      = exchangeItem.getLastModified () ;
                     localItemLastModified    = localItem.getLastModified    () ;

                     //
                     // because java.util.Date no save millisecond
                     // string rappresention of date is in HH:mm:ss
                     //
                     excItemLastModifiedStr   = String.valueOf(excItemLastModified)   ;
                     localItemLastModifiedStr = String.valueOf(localItemLastModified) ;

                     if (!excItemLastModifiedStr.equals(localItemLastModifiedStr)) {

                        updateRows.add(exchangeItem);
                        count++;

                     }

                 }

            }

        }

        return (Item[])updateRows.toArray(new Item[count]) ;

    }

    /**
     * get Item by id
     *
     * @param items array
     * @param id
     * @return <p>item</p> if find
     *         <p>null</p> if not find
     **/
    private Item getItemById(Item[] items,
                             String id) {

        Item item          = null;
        Item sequenceItem  = null;

        for (int i = 0, l = items.length; i < l; i++) {

            sequenceItem = items[i];

            if (sequenceItem.getId().equals(id)) {
                item = sequenceItem;
                break;
            }

        }

        return item;

    }

    /**
     *
     * check if item id is included
     * in ids from array of items
     *
     * @param items items array
     * @param id id to find
     * @return <p>true</p> if is included
     *         <p>false</p> if is not included
     **/
    private boolean isIdInArray(Item[] items,
                                String id) {

        boolean isFind = false;

        for (int i = 0, l = items.length; i < l; i++) {

            if (items[i].getId().equals(id)) {
                isFind = true;
                break;
            }

        }

        return isFind;

    }

    /**
     *
     * check if item is included
     * in array of items
     *
     * @param items items array
     * @param item item to find
     * @return <p>true</p> if is included
     *         <p>false</p> if is not included
     **/
    private boolean isItemInArray(Item[] items,
                                  Item   item) {

        return isIdInArray(items, item.getId());

    }


    /**
     * Return the device with the given deviceId
     * @param deviceId String
     * @return Sync4jDevice
     * @throws PersistentStoreException
     */
    private Sync4jDevice getDevice(String deviceId) throws PersistentStoreException {
        Sync4jDevice device = new Sync4jDevice(deviceId);
        PersistentStore store = Configuration.getConfiguration().getStore();
        store.read(device);
        return device;
    }


}
