/**
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.exchange.security;

import java.security.Principal;
import java.util.logging.Level;
import java.util.logging.Logger;

import sync4j.exchange.DataAccessException;

import sync4j.exchange.items.common.manager.ItemManager;

import sync4j.framework.core.Authentication;
import sync4j.framework.core.Cred;
import sync4j.framework.security.Officer;
import sync4j.framework.security.Sync4jPrincipal;
import sync4j.framework.server.store.*;
import sync4j.framework.tools.Base64;

import sync4j.server.config.Configuration;


/**
 * This is an implementation of the <i>Officier</i> interface.
 * It always authenticates and authorizes users and resource accesses.
 *
 * @version $Id: ExchangeOfficer.java,v 1.4 2006/02/13 13:39:37 fabius Exp $
 */
public class ExchangeOfficer implements Officer, java.io.Serializable {


    // ---------------------------------------------------------- Constants

    // ---------------------------------------------------------- Protected data

    protected Logger log = Logger.getLogger("source");

    protected PersistentStore ps = null;

    // ------------------------------------------------------------- Public data

    // -------------------------------------------------------------- Properties

    /**
     * Has the last login failed for incorrect login/password?
     */
    private boolean loginFailed = false;

    public boolean isLoginFailed() {
        return loginFailed;
    }

    /**
     * Has the last login failed for expired temporary login?
     */
    private boolean loginExpired = false;

    /**
     * Which type of authetication does impose the server?
     */
    private String clientAuth = Cred.AUTH_TYPE_BASIC;

    public String getClientAuth() {
        return this.clientAuth;
    }

    public void setClientAuth(String clientAuth) {
        this.clientAuth = clientAuth;
    }

    /**
     * Which type of authetication use the server to send his credential?
     */
    private String serverAuth = Cred.AUTH_TYPE_BASIC;

    public String getServerAuth() {
        return this.serverAuth;
    }

    public void setServerAuth(String serverAuth) {
        this.serverAuth = serverAuth;
    }

    /**
     * Exchange host
     */
    private String exchangeHost = null;

    public String getExchangeHost() {
        return this.exchangeHost;
    }

    public void setExchangeHost(String exchangeHost) {
        this.exchangeHost = exchangeHost;
    }

    /**
     * Exchange port
     */
    private int exchangePort = 0;

    public int getExchangePort() {
        return this.exchangePort;
    }

    public void setExchangePort(int exchangePort) {
        this.exchangePort = exchangePort;
    }

    /**
     * Exchange folder
     */
    private String exchangeFolder = null;

    public String getExchangeFolder() {
        return this.exchangeFolder;
    }

    public void setExchangeFolder(String exchangeFolder) {
        this.exchangeFolder = exchangeFolder;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Authenticates a credential.
     *
     * @param credential the credential to be authenticated
     *
     * @return true if the credential is autenticated, false otherwise
     */
    public boolean authenticate(Cred credential) {

        if (log.isLoggable(Level.FINEST)) {
            log.finest(" ExchangeOfficer authenticate() Start");
        }

        Configuration config = Configuration.getConfiguration();
        ps = config.getStore();

        String type = credential.getType();

        if ((Cred.AUTH_TYPE_BASIC).equals(type)) {
            return authenticateBasicCredential(credential);
        }
        return false;
    }


    /**
     * Authorizes a resource.
     *
     * @param principal the requesting entity
     * @param resource the name (or the identifier) of the resource to be authorized
     *
     * @return true if the credential is authorized to access the resource, false
     *         otherwise
     */
    public boolean authorize(Principal principal, String resource) {
        return true;
    }

    /**
     * Un-authenticates a credential.
     * Do nothing. In the current implementation, the authentication is
     * discarde as soon as the LoginContext is garbage collected.
     *
     * @param credential the credential to be unauthenticated
     */
    public void unAuthenticate(Cred credential) {
    }

    /**
     * @see sync4j.framework.security.Officer
     *
     * @return boolean
     */
    public boolean isAccountExpired() {
        return loginExpired;
    }

    // --------------------------------------------------------- Private Methods

    /**
     *
     */
    private boolean authenticateBasicCredential(Cred credential) {

        String username = null, password = null;
        Sync4jPrincipal principal = null;

        Authentication auth = credential.getAuthentication();
        String deviceId     = auth.getDeviceId();
        String credentials  = auth.getData();
        String userpwd      = new String(Base64.decode(auth.getData()));

        int p = userpwd.indexOf(':');

        if (p == -1) {
            username = userpwd;
            password = "";
        } else {
            username = (p > 0) ? userpwd.substring(0, p) : "";
            password = (p == (userpwd.length() - 1)) ? "" :
                       userpwd.substring(p + 1);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Username: "    + username    );
            log.finest("Credentials: " + credentials );
        }

        //
        // Verify that exist the principal for the
        // specify deviceId and username
        //
        boolean principalFound = false;
        try {
            principal = new Sync4jPrincipal(username, deviceId);
            ps.read(principal);
            credential.getAuthentication().setPrincipalId(principal.getId());
            principalFound = true;
        } catch (NotFoundException nfe) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Principal for " + username + ":" + deviceId +
                           " not found");
            }
        } catch (PersistentStoreException e) {
            log.severe("Error reading principal: " + e);
            log.throwing(getClass().getName(), "authenticateBasicCredential", e);
            return false;
        }

        //
        // Verify these credentials on Exchange
        //
        boolean isExchangeUser = checkExchangeCredentials(
            exchangeHost,
            exchangePort,
            exchangeFolder,
            username,
            credentials
        );

        if (!isExchangeUser) {
            //
            // User not authenticate
            //
            return false;
        }

        if (!principalFound) {
            //
            // We must create a new principal
            //
            principal = new Sync4jPrincipal(username, deviceId);
            try {
                ps.store(principal);
            } catch (PersistentStoreException ex) {
                ex.printStackTrace();
                log.severe("Error creating new principal: " + ex);
                log.throwing(getClass().getName(),
                             "authenticateBasicCredential", ex);
                return false;
            }

            credential.getAuthentication().setPrincipalId(principal.getId());
        }

        return true;
    }

    /**
     * Checks if the given login and password are valid
     * credentials on exchange server
     *
     *
     * @param server String
     * @param port int
     * @param login String
     * @param password String
     * @param version String
     * @param appName String
     *
     *
     * @return boolean
     */
    private boolean checkExchangeCredentials(String exchangeHost   ,
                                             int    exchangePort   ,
                                             String exchangeFolder ,
                                             String userName       ,
                                             String credentials) {
        int status = 0;

        if (exchangeHost == null || exchangeHost.equals("")) {

            String message = "'Exchange host' is null or empty. You have to set " +
                             "the Exchange server host in the file: " +
                             "Sync4j/config/sync4j/server/security/ExchangeOfficer.xml";

            if (log.isLoggable(Level.FINEST)) {
                log.finest(message);
            }

            throw new IllegalStateException(message);
        }


        if (exchangeFolder == null || exchangeFolder.equals("")) {

            String message = "'Exchange folder' is null or empty. You have to set " +
                             "the Exchange folder folder in the file: " +
                             "Sync4j/config/sync4j/server/security/ExchangeOfficer.xml";

            if (log.isLoggable(Level.FINEST)) {
                log.finest(message);
            }

            throw new IllegalStateException(message);
        }

        if (exchangeFolder.indexOf("/") == - 1) {

            String message = "'Exchange folder' must be contain '/'. You have to set " +
                             "the Exchange folder host in the file: " +
                             "Sync4j/config/sync4j/server/security/ExchangeOfficer.xml";

            if (log.isLoggable(Level.FINEST)) {
                log.finest(message);
            }

            throw new IllegalStateException(message);
        }

        String exchangeServerName
                        = exchangeFolder.
                            substring(0,
                                      exchangeFolder.indexOf("/"));

        boolean check = false;

        try {

            ItemManager im = new ItemManager (exchangeHost,
                                              exchangePort);

            status = im.getExchangeAccessStatus(exchangeServerName ,
                                                userName           ,
                                                credentials        );

            if (status == 401 || status == 406) {
                check = false;
            } else {
                check = true;
            }

        } catch (DataAccessException dae) {
            log.severe("Error getting connection: " + dae);
            log.throwing(getClass().getName(), "check Exchange Credentials",
                         dae);
        }
        return check;
    }

}
